import typing as tp  # NOQA

from chainer.functions.connection import linear
from chainer import initializers
from chainer import link
from chainer import types  # NOQA
from chainer import utils
from chainer import variable


class Linear(link.Link):

    """Linear layer (a.k.a.\\  fully-connected layer).

    This is a link that wraps the :func:`~chainer.functions.linear` function,
    and holds a weight matrix ``W`` and optionally a bias vector ``b`` as
    parameters.

    If ``initialW`` is left to the default value of ``None``, the weight matrix
    ``W`` is initialized with i.i.d. Gaussian samples, each of which has zero
    mean and deviation :math:`\\sqrt{1/\\text{in_size}}`. The bias vector ``b``
    is of size ``out_size``. If the ``initial_bias`` is to left the default
    value of ``None``, each element is initialized as zero.  If the ``nobias``
    argument is set to ``True``, then this link does not hold a bias vector.

    Args:
        in_size (int or None): Dimension of input vectors. If unspecified or
            ``None``, parameter initialization will be deferred until the
            first forward data pass at which time the size will be determined.
        out_size (int): Dimension of output vectors. If only one value is
            passed for ``in_size`` and ``out_size``, that value will be used
            for the ``out_size`` dimension.
        nobias (bool): If ``True``, then this function does not use the bias.
        initialW (:ref:`initializer <initializer>`): Initializer to initialize
            the weight. When it is :class:`numpy.ndarray`,
            its ``ndim`` should be 2. If ``initialW`` is ``None``, then the
            weights are initialized with i.i.d. Gaussian samples, each of which
            has zero mean and deviation :math:`\\sqrt{1/\\text{in_size}}`.
        initial_bias (:ref:`initializer <initializer>`): Initializer to
            initialize the bias. If ``None``, the bias will be initialized to
            zero. When it is :class:`numpy.ndarray`, its ``ndim`` should be 1.
    .. seealso:: :func:`~chainer.functions.linear`

    Attributes:
        W (~chainer.Variable): Weight parameter.
        b (~chainer.Variable): Bias parameter.

    .. admonition:: Example

        There are several ways to make a Linear link.

        Define an input vector ``x`` as:

        >>> x = np.array([[0, 1, 2, 3, 4]], np.float32)

        1. Give the first two arguments explicitly:

            Those numbers are considered as the input size and the output size.

            >>> l = L.Linear(5, 10)
            >>> y = l(x)
            >>> y.shape
            (1, 10)

        2. Omit ``in_size`` (give the output size only as the first argument)
           or fill it with ``None``:

            In this case, the size of second axis of ``x`` is used as the
            input size. So the below two cases are the same.

            >>> l = L.Linear(10)
            >>> y = l(x)
            >>> y.shape
            (1, 10)

            >>> l = L.Linear(None, 10)
            >>> y = l(x)
            >>> y.shape
            (1, 10)

            When you omit the first argument, you need to specify the other
            subsequent arguments from ``nobias`` as keyword arguments. So the
            below two cases are the same.

            >>> l = L.Linear(None, 10, False, None, 0)
            >>> y = l(x)
            >>> y.shape
            (1, 10)

            >>> l = L.Linear(10, nobias=False, initialW=None, initial_bias=0)
            >>> y = l(x)
            >>> y.shape
            (1, 10)

    """

    def __init__(
            self,
            in_size: tp.Optional[int],
            out_size: tp.Optional[int] = None,
            nobias: bool = False,
            initialW: tp.Optional[types.InitializerSpec] = None,
            initial_bias: tp.Optional[types.InitializerSpec] = None
    ) -> None:
        super(Linear, self).__init__()

        if out_size is None:
            in_size, out_size = None, in_size
        self.in_size = in_size
        self.out_size = out_size

        with self.init_scope():
            W_initializer = initializers._get_initializer(initialW)
            self.W = variable.Parameter(W_initializer)  # type: variable.Variable  # NOQA
            if in_size is not None:
                self._initialize_params(in_size)

            if nobias:
                self.b = None  # type: tp.Optional[variable.Variable]
            else:
                if initial_bias is None:
                    initial_bias = 0
                bias_initializer = initializers._get_initializer(initial_bias)
                self.b = variable.Parameter(bias_initializer, out_size)

    def _initialize_params(self, in_size: int) -> None:
        self.W.initialize((self.out_size, in_size))  # type: ignore

    @classmethod
    def from_params(cls, W, b=None, nobias=False):
        """Initialize a :class:`~chainer.links.Linear` with given parameters.

        This method uses ``W`` and optional ``b`` to initialize a linear layer.

        Args:
            W (:class:`~chainer.Variable` or :ref:`ndarray`):
                The weight parameter.
            b (:class:`~chainer.Variable`, :ref:`ndarray`, or ``None``):
                The bias parameter.
            nobias (bool): If ``True``, the argument of ``b`` is ignored
                in spite of whether it's given or not.
        """
        out_size, in_size = W.shape
        if b is not None:
            if out_size != b.size:
                raise ValueError('`out_size` does not match the size of `b`')
        link = cls(
            in_size, out_size, nobias,
            initialW=variable.as_array(W), initial_bias=variable.as_array(b))
        return link

    @property
    def printable_specs(self):
        specs = [
            ('in_size', self.in_size),
            ('out_size', self.out_size),
            ('nobias', self.b is None),
        ]
        for spec in specs:
            yield spec

    def forward(
            self,
            x: variable.Variable,
            n_batch_axes: int = 1
    ) -> variable.Variable:
        """Applies the linear layer.

        Args:
            x (~chainer.Variable): Batch of input vectors.
            n_batch_axes (int): The number of batch axes. The default is 1. The
                input variable is reshaped into
                (:math:`{\\rm n\\_batch\\_axes} + 1`)-dimensional tensor.
                This should be greater than 0.

        Returns:
            ~chainer.Variable: Output of the linear layer.

        """
        if self.W.array is None:
            in_size = utils.size_of_shape(x.shape[n_batch_axes:])
            self._initialize_params(in_size)
        return linear.linear(x, self.W, self.b, n_batch_axes=n_batch_axes)



import json
from os import path
import warnings

import numpy
import six

from chainer import reporter
from chainer import serializer as serializer_module
from chainer.training import extension
from chainer.training import trigger as trigger_module
from chainer.utils import argument


_available = None


def _try_import_matplotlib():
    global matplotlib, _available
    try:
        import matplotlib  # NOQA
        _available = True
    except (ImportError, TypeError):
        _available = False


def _check_available():
    if _available is None:
        _try_import_matplotlib()

    if not _available:
        warnings.warn('matplotlib is not installed on your environment, '
                      'so nothing will be plotted at this time. '
                      'Please install matplotlib to plot figures.\n\n'
                      '  $ pip install matplotlib\n')


class PlotReport(extension.Extension):

    """__init__(\
y_keys, x_key='iteration', trigger=(1, 'epoch'), postprocess=None, \
filename='plot.png', marker='x', grid=True)

    Trainer extension to output plots.

    This extension accumulates the observations of the trainer to
    :class:`~chainer.DictSummary` at a regular interval specified by a supplied
    trigger, and plot a graph with using them.

    There are two triggers to handle this extension. One is the trigger to
    invoke this extension, which is used to handle the timing of accumulating
    the results. It is set to ``1, 'iteration'`` by default. The other is the
    trigger to determine when to emit the result. When this trigger returns
    True, this extension appends the summary of accumulated values to the list
    of past summaries, and writes the list to the log file. Then, this
    extension makes a new fresh summary object which is used until the next
    time that the trigger fires.

    It also adds ``'epoch'`` and ``'iteration'`` entries to each result
    dictionary, which are the epoch and iteration counts at the output.

    .. warning::

        If your environment needs to specify a backend of matplotlib
        explicitly, please call ``matplotlib.use`` before calling
        ``trainer.run``. For example:

        .. code-block:: python

            import matplotlib
            matplotlib.use('Agg')

            trainer.extend(
                extensions.PlotReport(['main/loss', 'validation/main/loss'],
                                      'epoch', filename='loss.png'))
            trainer.run()

        Then, once one of instances of this extension is called,
        ``matplotlib.use`` will have no effect.

    For the details, please see here:
    https://matplotlib.org/faq/usage_faq.html#what-is-a-backend

    Args:
        y_keys (iterable of strs): Keys of values regarded as y. If this is
            ``None``, nothing is output to the graph.
        x_key (str): Keys of values regarded as x. The default value is
            'iteration'.
        trigger: Trigger that decides when to aggregate the result and output
            the values. This is distinct from the trigger of this extension
            itself. If it is a tuple in the form ``<int>, 'epoch'`` or ``<int>,
            'iteration'``, it is passed to :class:`IntervalTrigger`.
        postprocess: Callback to postprocess the result dictionaries. Figure
            object, Axes object, and all plot data are passed to this callback
            in this order. This callback can modify the figure.
        filename (str): Name of the figure file under the output directory.
            It can be a format string.
            For historical reasons ``file_name`` is also accepted as an alias
            of this argument.
        marker (str): The marker used to plot the graph. Default is ``'x'``. If
            ``None`` is given, it draws with no markers.
        grid (bool): If ``True``, set the axis grid on.
            The default value is ``True``.

    """

    def __init__(self, y_keys, x_key='iteration', trigger=(1, 'epoch'),
                 postprocess=None, filename=None, marker='x',
                 grid=True, **kwargs):

        file_name, = argument.parse_kwargs(kwargs, ('file_name', 'plot.png'))
        if filename is None:
            filename = file_name
        del file_name  # avoid accidental use

        _check_available()

        self._x_key = x_key
        if isinstance(y_keys, str):
            y_keys = (y_keys,)

        self._y_keys = y_keys
        self._trigger = trigger_module.get_trigger(trigger)
        self._file_name = filename
        self._marker = marker
        self._grid = grid
        self._postprocess = postprocess
        self._init_summary()
        self._data = {k: [] for k in y_keys}

    @staticmethod
    def available():
        _check_available()
        return _available

    def __call__(self, trainer):
        if self.available():
            # Dynamically import pyplot to call matplotlib.use()
            # after importing chainer.training.extensions
            import matplotlib.pyplot as plt
        else:
            return

        keys = self._y_keys
        observation = trainer.observation
        summary = self._summary

        if keys is None:
            summary.add(observation)
        else:
            summary.add({k: observation[k] for k in keys if k in observation})

        if trainer.is_before_training or self._trigger(trainer):
            stats = self._summary.compute_mean()
            stats_cpu = {}
            for name, value in six.iteritems(stats):
                stats_cpu[name] = float(value)  # copy to CPU

            updater = trainer.updater
            stats_cpu['epoch'] = updater.epoch
            stats_cpu['iteration'] = updater.iteration
            x = stats_cpu[self._x_key]
            data = self._data

            for k in keys:
                if k in stats_cpu:
                    data[k].append((x, stats_cpu[k]))

            f = plt.figure()
            a = f.add_subplot(111)
            a.set_xlabel(self._x_key)
            if self._grid:
                a.grid()

            for k in keys:
                xy = data[k]
                if len(xy) == 0:
                    continue

                xy = numpy.array(xy)
                a.plot(xy[:, 0], xy[:, 1], marker=self._marker, label=k)

            if a.has_data():
                if self._postprocess is not None:
                    self._postprocess(f, a, summary)
                l = a.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
                f.savefig(path.join(trainer.out, self._file_name),
                          bbox_extra_artists=(l,), bbox_inches='tight')

            plt.close()
            self._init_summary()

    def serialize(self, serializer):
        if isinstance(serializer, serializer_module.Serializer):
            serializer('_plot_{}'.format(self._file_name),
                       json.dumps(self._data))

        else:
            self._data = json.loads(
                serializer('_plot_{}'.format(self._file_name), ''))

    def _init_summary(self):
        self._summary = reporter.DictSummary()



import chainer
from chainer import function_node
from chainer.utils import type_check


# TODO(niboshi): memory layouts and conversions should better be implemented
# with polymorphism instead of ad-hoc conditions.

CUDNN_CHANNEL_FIRST_X = None
CUDNN_CHANNEL_LAST_X = 'CUDNN_CHANNEL_LAST_X'
CUDNN_CHANNEL_FIRST_W = None
CUDNN_CHANNEL_LAST_W = 'CUDNN_CHANNEL_LAST_W'


class _Unspecified:
    def __repr__(self):
        return '<unspecified>'


_unspecified = _Unspecified()


def get_raw_shape(arr_or_var):
    if isinstance(arr_or_var, chainer.Variable):
        arr = arr_or_var._data[0]
    else:
        arr = arr_or_var
    return arr.shape


def get_semantic_shape(arr_or_var, *, assumed_layout=_unspecified):
    if not isinstance(arr_or_var, chainer.Variable):
        # array
        shape = arr_or_var.shape
        if assumed_layout is not _unspecified:
            shape = _transpose_shape(shape, assumed_layout, None)
        return shape

    # variable
    if assumed_layout is not _unspecified:
        # TODO(niboshi): Raise exception
        assert arr_or_var.layout == assumed_layout
    return arr_or_var.shape


def _transpose_array(arr, src_layout, dst_layout):
    trans = _get_layout_transpose_axes(arr.ndim, src_layout, dst_layout)
    if trans is None:
        return arr
    return arr.transpose(*trans)


def _transpose_shape(shape, src_layout, dst_layout):
    trans = _get_layout_transpose_axes(len(shape), src_layout, dst_layout)
    if trans is None:
        return shape
    return tuple([shape[i] for i in trans])


def _get_layout_transpose_axes(ndim, src_layout, dst_layout, inverse=False):
    # None: no transposition is required.

    if src_layout == dst_layout:
        return None

    if dst_layout == CUDNN_CHANNEL_LAST_X:
        assert ndim >= 3
        assert src_layout == CUDNN_CHANNEL_FIRST_X
        trans = (0,) + tuple(range(2, ndim)) + (1,)

    elif dst_layout == CUDNN_CHANNEL_LAST_W:
        assert ndim >= 3
        assert src_layout == CUDNN_CHANNEL_FIRST_W
        trans = (0,) + tuple(range(2, ndim)) + (1,)

    elif src_layout == CUDNN_CHANNEL_LAST_X:
        assert ndim >= 3
        assert dst_layout == CUDNN_CHANNEL_FIRST_X
        trans = (0, ndim-1) + tuple(range(1, ndim-1))

    elif src_layout == CUDNN_CHANNEL_LAST_W:
        assert ndim >= 3
        assert dst_layout == CUDNN_CHANNEL_FIRST_W
        trans = (0, ndim-1) + tuple(range(1, ndim-1))

    else:
        raise ValueError(
            'Unknown layout conversion: from \'{}\' to \'{}\''.format(
                src_layout, dst_layout))

    if inverse:
        t = [None] * ndim
        for i, n in enumerate(trans):
            t[n] = i
        trans = tuple(t)

    # Postconditions:
    # - assert isinstance(trans, tuple)
    # - assert len(trans) == ndim
    # - assert all([i in trans for i in range(ndim)])
    return trans


class AsLayout(function_node.FunctionNode):
    """Permute the dimensions of an array."""

    axes = None
    in_layout = None

    def __init__(self, layout):
        self.out_layout = layout

    def check_type_forward(self, in_types):
        type_check.expect(in_types.size() == 1,)

    def check_layout_forward(self, inputs):
        x, = inputs
        self.axes = _get_layout_transpose_axes(
            x.ndim, x.layout, self.out_layout)
        self.in_layout = x.layout

    @property
    def label(self):
        return 'AsLayout'

    def forward_chainerx(self, inputs):
        # TODO(niboshi): Add support for this
        raise RuntimeError(
            'Non-standard memory layouts are not supported for chainerx.')

    def forward(self, inputs):
        x, = inputs
        axes = self.axes
        self.output_layouts = (self.out_layout,)
        if axes is None:
            return x
        return x.transpose(axes),

    def backward(self, indexes, grad_outputs):
        return AsLayout(self.in_layout).apply(grad_outputs)



import chainer
from chainer import function_node
from chainer.utils import type_check
import chainerx


class Cholesky(function_node.FunctionNode):

    @property
    def label(self):
        return 'cholesky'

    def check_type_forward(self, in_types):
        type_check._argname(in_types, ('a', ))
        a_type, = in_types

        type_check.expect(
            a_type.dtype.kind == 'f',
            a_type.ndim == 2,
        )

    def forward(self, inputs):
        a, = inputs
        self.retain_outputs((0,))
        xp = chainer.backend.get_array_module(a)
        return xp.linalg.cholesky(a),

    def forward_chainerx(self, inputs):
        return chainerx.linalg.cholesky(*inputs),

    def backward(self, indexes, grad_outputs):
        gy, = grad_outputs
        xp = chainer.backend.get_array_module(gy)
        y, = self.get_retained_outputs()
        n = y.shape[0]
        dtype = y.dtype

        F = chainer.functions
        y_inv = F.inv(y)
        mask = xp.tri(n, dtype=dtype) - 0.5 * xp.eye(n, dtype=dtype)
        phi = mask * F.matmul(y, gy, transa=True)
        s = F.matmul(F.matmul(y_inv, phi, transa=True), y_inv)
        gx = 0.5 * (s + s.T)
        return gx,


def cholesky(a):
    """Cholesky Decomposition

    Args:
        a (:class:`~chainer.Variable` or :ref:`ndarray`): Input variable.

    Returns:
        ~chainer.Variable: Output variable.
    """
    return Cholesky().apply((a,))[0]



import chainer

from onnx_chainer.functions.opset_version import support
from onnx_chainer import onnx_helper


@support((1, 6, 7))
def convert_n_step_gru(func, opset_version, input_names, output_names,
                       context):
    n_layers, dropout_ratio, hx, ws, bs, xs = func.args
    assert n_layers >= 1
    hidden_size = hx.shape[2]

    gb = onnx_helper.GraphBuilder()

    hx_name = input_names[0]
    offset = 1
    ws_names = [[input_names[offset + i * 6 + j] for j in range(6)]
                for i in range(n_layers)]
    offset += 6 * n_layers
    bs_names = [[input_names[offset + i * 6 + j] for j in range(6)]
                for i in range(n_layers)]
    offset += 6 * n_layers
    xs_names = input_names[offset:]

    split_outs = gb.op('Split', [hx_name], num_outputs=n_layers, axis=0)
    if n_layers == 1:
        split_outs = [split_outs]
    # Removing layer dimention and adding num_directions cancels each other
    hx_names = split_outs

    hy_name, ys_name_list = \
        func.reconstruct_return_value(output_names)

    y_name = None
    hy_names = []

    for layer in range(n_layers):
        if layer == 0:
            # X; shape: (seq_length, batch_size, input_size)
            x_name = gb.op(
                'Concat',
                [gb.op('Unsqueeze', [name], axes=[0]) for name in xs_names],
                axis=0)
        else:
            if opset_version >= 7:
                x_name = gb.op('Dropout', [y_name], ratio=dropout_ratio)
            elif opset_version >= 6:
                x_name = gb.op('Dropout', [y_name], ratio=dropout_ratio,
                               is_test=0 if chainer.config.train else 1)
            else:
                x_name = gb.op('Dropout', [y_name], ratio=dropout_ratio,
                               is_test=0 if chainer.config.train else 1,
                               consumed_inputs=[1])

            # remove num_directions dimention
            x_name = gb.op('Squeeze', [x_name], axes=[1])

        w = ws_names[layer]
        b = bs_names[layer]

        # W[zrh]; shape: (num_directions, 3*hidden_size, input_size)
        w_name = gb.op(
            'Unsqueeze',
            [gb.op('Concat', [w[1], w[0], w[2]], axis=0)],
            axes=[0])
        # R[zrh]; shape: (num_directions, 3*hidden_size, input_size)
        r_name = gb.op(
            'Unsqueeze',
            [gb.op('Concat', [w[4], w[3], w[5]], axis=0)],
            axes=[0])
        # Wb[zrh], Rb[zrh]; shape: (num_directions, 6*hidden_size)
        b_name = gb.op(
            'Unsqueeze',
            [gb.op('Concat', [b[1], b[0], b[2], b[4], b[3], b[5]], axis=0)],
            axes=[0])

        # Y; shape: (seq_length, num_directions, batch_size, hidden_size)
        # Y_h; shape: (num_directions, batch_size, hidden_size)
        y_name, hy_name_ = gb.op(
            'GRU',
            (x_name, w_name, r_name, b_name, "", hx_names[layer]),
            hidden_size=hidden_size,
            linear_before_reset=1,
            num_outputs=2)
        hy_names.append(hy_name_)

    split_outs = gb.op(
        'Split',
        # remove num_directions dimention
        [gb.op('Squeeze', [y_name], axes=[1])],
        num_outputs=len(ys_name_list), axis=0)
    if len(ys_name_list) == 1:
        split_outs = [split_outs]
    for i, node_name in enumerate(split_outs):
        # remove seq_length dimention
        gb.op_output_named('Squeeze', [node_name], [ys_name_list[i]], axes=[0])

    # Removal of num_directions and new dimention for concatenation cancel
    # each other.
    gb.op_output_named('Concat', hy_names, [hy_name], axis=0)

    return gb.nodes()



from chainermn.extensions.allreduce_persistent import AllreducePersistent  # NOQA
from chainermn.extensions.checkpoint import create_multi_node_checkpointer  # NOQA
from chainermn.extensions.multi_node_evaluator import create_multi_node_evaluator  # NOQA
from chainermn.extensions.multi_node_evaluator import GenericMultiNodeEvaluator  # NOQA
from chainermn.extensions._multi_node_snapshot import multi_node_snapshot  # NOQA
from chainermn.extensions.observation_aggregator import ObservationAggregator  # NOQA
from chainermn.extensions.multi_node_early_stopping_trigger import MultiNodeEarlyStoppingTrigger  # NOQA



import numpy

import chainer
from chainer import backend
from chainer import function_node
from chainer.utils import type_check


def _get_intermediate_dtype(dtype):
    # Returns the dtype for intermediate calculation.
    # For float16 input, float32 is used.
    # Otherwise the same dtype as the parameter is used.
    if dtype == numpy.float16:
        return numpy.float32
    return dtype


class MeanAbsoluteError(function_node.FunctionNode):

    """Mean absolute error function."""

    def check_type_forward(self, in_types):
        type_check._argname(in_types, ('x0', 'x1'))
        type_check.expect(
            in_types[0].dtype.kind == 'f',
            in_types[0].dtype == in_types[1].dtype,
            in_types[0].shape == in_types[1].shape
        )

    def forward_cpu(self, inputs):
        x0, x1 = inputs
        self.diff = x0 - x1
        orig_dtype = self.diff.dtype
        dtype = _get_intermediate_dtype(orig_dtype)
        diff = self.diff.ravel().astype(dtype, copy=False)
        return numpy.array(abs(diff).sum() / diff.size, dtype=orig_dtype),

    def forward_gpu(self, inputs):
        x0, x1 = inputs
        self.diff = x0 - x1
        orig_dtype = self.diff.dtype
        dtype = _get_intermediate_dtype(orig_dtype)
        diff = self.diff.ravel().astype(dtype, copy=False)
        return (abs(diff).sum() / diff.dtype.type(diff.size)).astype(
            orig_dtype, copy=False),

    def backward(self, indexes, grad_outputs):
        gy, = grad_outputs
        coeff = gy * gy.data.dtype.type(1. / self.diff.size)
        coeff = chainer.functions.broadcast_to(coeff, self.diff.shape)
        gx0 = coeff * backend.get_array_module(gy.data).sign(self.diff)
        return gx0, -gx0


def mean_absolute_error(x0, x1):
    """Mean absolute error function.

    The function computes the mean absolute error between two variables. The
    mean is taken over the minibatch. Args ``x0`` and ``x1`` must have the
    same dimensions. This function first calculates the absolute value
    differences between the corresponding elements in x0 and x1, and then
    returns the mean of those differences.

    Args:
        x0 (:class:`~chainer.Variable` or :ref:`ndarray`): Input variable.
        x1 (:class:`~chainer.Variable` or :ref:`ndarray`): Input variable.

    Returns:
        ~chainer.Variable:
            A variable holding an array representing the mean absolute
            error of two inputs.

    .. admonition:: Example

        1D array examples:

        >>> x = np.array([1, 2, 3]).astype(np.float32)
        >>> y = np.array([0, 0, 0]).astype(np.float32)
        >>> F.mean_absolute_error(x, y)
        variable(2.)
        >>> x = np.array([1, 2, 3, 4, 5, 6]).astype(np.float32)
        >>> y = np.array([7, 8, 9, 10, 11, 12]).astype(np.float32)
        >>> F.mean_absolute_error(x, y)
        variable(6.)

        2D array example:

        In this example, there are 4 elements, and thus 4 errors
        >>> x = np.array([[1, 2], [3, 4]]).astype(np.float32)
        >>> y = np.array([[8, 8], [8, 8]]).astype(np.float32)
        >>> F.mean_absolute_error(x, y)
        variable(5.5)

        3D array example:

        In this example, there are 8 elements, and thus 8 errors
        >>> x = np.reshape(np.array([1, 2, 3, 4, 5, 6, 7, 8]), (2, 2, 2))
        >>> y = np.reshape(np.array([8, 8, 8, 8, 8, 8, 8, 8]), (2, 2, 2))
        >>> x = x.astype(np.float32)
        >>> y = y.astype(np.float32)
        >>> F.mean_absolute_error(x, y)
        variable(3.5)

    """
    return MeanAbsoluteError().apply((x0, x1))[0]



import sys

import numpy
import six

from chainer.backends import _chainerx
from chainer.backends import _cpu
from chainer.backends import cuda
from chainer.backends import intel64
from chainer import serializer
import chainerx


try:
    import h5py
    _available = True
except ImportError:
    _available = False


def _check_available():
    if not _available:
        msg = '''h5py is not installed on your environment.
Please install h5py to activate hdf5 serializers.

  $ pip install h5py'''
        raise RuntimeError(msg)


class HDF5Serializer(serializer.Serializer):

    """Serializer for HDF5 format.

    This is the standard serializer in Chainer. The chain hierarchy is simply
    mapped to HDF5 hierarchical groups.

    Args:
        group (h5py.Group): The group that this serializer represents.
        compression (int): Gzip compression level.

    """

    def __init__(self, group, compression=4):
        _check_available()

        self.group = group
        self.compression = compression

    def __getitem__(self, key):
        name = self.group.name + '/' + key
        return HDF5Serializer(self.group.require_group(name), self.compression)

    def __call__(self, key, value):
        if value is None:
            # use Empty to represent None
            if h5py.version.version_tuple < (2, 7, 0):
                raise RuntimeError(
                    'h5py>=2.7.0 is required to serialize None.')
            arr = h5py.Empty('f')
            compression = None
        else:
            arr = _cpu._to_cpu(value)
            compression = None if arr.size <= 1 else self.compression
        self.group.create_dataset(key, data=arr, compression=compression)
        return value


def save_hdf5(filename, obj, compression=4):
    """Saves an object to the file in HDF5 format.

    This is a short-cut function to save only one object into an HDF5 file. If
    you want to save multiple objects to one HDF5 file, use
    :class:`HDF5Serializer` directly by passing appropriate :class:`h5py.Group`
    objects.

    Args:
        filename (str): Target file name.
        obj: Object to be serialized. It must support serialization protocol.
            If it is a dictionary object, the serialization will be skipped.
        compression (int): Gzip compression level.

    .. note::
        Currently :func:`save_hdf5` only supports writing to an actual file on
        file system due to a limitation of HD5F library.
        See `h5py/h5py#687 <https://github.com/h5py/h5py/issues/687>`_ for
        details.

    .. seealso::
        :func:`chainer.serializers.load_hdf5`

    """
    _check_available()
    with h5py.File(filename, 'w') as f:
        if isinstance(obj, dict):
            for key, value in obj.items():
                key = '/' + key.lstrip('/')
                arr = numpy.asarray(value)
                compression = None if arr.size <= 1 else compression
                try:
                    f.create_dataset(key, data=arr, compression=compression)
                except TypeError:
                    sys.stderr.write(
                        'A key named "{}" is unable to save in HDF5 format.\n')
                    # In Chainer, LogReport extension and PlotReport extension
                    # are # unable to save in HDF5 format. These extensions
                    # have a data type `numpy.dtype('O')` which is not
                    # supported by h5py.
                    six.reraise(*sys.exec_info())
        else:
            s = HDF5Serializer(f, compression=compression)
            s.save(obj)


class HDF5Deserializer(serializer.Deserializer):

    """Deserializer for HDF5 format.

    This is the standard deserializer in Chainer. This deserializer can be used
    to read an object serialized by :class:`HDF5Serializer`.

    Args:
        group (h5py.Group): The group that the deserialization starts from.
        strict (bool): If ``True``, the deserializer raises an error when an
            expected value is not found in the given HDF5 file. Otherwise,
            it ignores the value and skip deserialization.

    """

    def __init__(self, group, strict=True):
        _check_available()
        self.group = group
        self.strict = strict

    def __getitem__(self, key):
        name = self.group.name + '/' + key
        try:
            group = self.group.require_group(name)
        except ValueError:
            # require_group raises ValueError if there does not exist
            # the given group and the file is read mode.
            group = None
        return HDF5Deserializer(group, strict=self.strict)

    def __call__(self, key, value):
        if self.group is None:
            if not self.strict:
                return value
            else:
                raise ValueError('Inexistent group is specified')
        if not self.strict and key not in self.group:
            return value

        dataset = self.group[key]
        if dataset.shape is None:  # Empty
            return None
        if value is None:
            return numpy.asarray(dataset)
        if isinstance(value, chainerx.ndarray):
            value[...] = _chainerx._array_to_chainerx(
                numpy.asarray(dataset), value.device)
        elif isinstance(value, numpy.ndarray):
            dataset.read_direct(value)
        elif isinstance(value, cuda.ndarray):
            value.set(numpy.asarray(dataset, dtype=value.dtype))
        elif isinstance(value, intel64.mdarray):
            intel64.ideep.basic_copyto(value, numpy.asarray(dataset))
        else:
            value = type(value)(numpy.asarray(dataset))
        return value


def load_hdf5(filename, obj):
    """Loads an object from the file in HDF5 format.

    This is a short-cut function to load from an HDF5 file that contains only
    one object. If you want to load multiple objects from one HDF5 file, use
    :class:`HDF5Deserializer` directly by passing appropriate
    :class:`h5py.Group` objects.

    Args:
        filename (str): Name of the file to be loaded.
        obj: Object to be deserialized. It must support serialization protocol.

    .. note::
        Currently :func:`load_hdf5` only supports loading an actual file on
        file system due to a limitation of HD5F library.
        See `h5py/h5py#687 <https://github.com/h5py/h5py/issues/687>`_ for
        details.

    .. seealso::
        :func:`chainer.serializers.save_hdf5`

    """
    _check_available()
    with h5py.File(filename, 'r') as f:
        d = HDF5Deserializer(f)
        d.load(obj)



from contextlib import contextmanager

import chainer.functions as F

from onnx_chainer import functions
from onnx_chainer.functions.converter import FunctionConverter
from onnx_chainer.replace_func import fake_as_funcnode


_supported_function_node_set = {
    # Activation
    'ClippedReLU',
    'ELU',
    'HardSigmoid',
    'LeakyReLU',
    'LogSoftmax',
    'PReLUFunction',
    'ReLU',
    'Selu',
    'Sigmoid',
    'Softmax',
    'Softplus',
    'Tanh',

    # Array
    'Cast',
    'Concat',
    'Copy',
    'Depth2Space',
    'Dstack',
    'ExpandDims',
    'GetItem',
    'Hstack',
    'Moveaxis',
    'Pad',
    'Permutate',
    'Repeat',
    'Reshape',
    'ResizeImages',
    'Rollaxis',
    'SelectItem',
    'Separate',
    'Shape',
    'Space2Depth',
    'SplitAxis',
    'Squeeze',
    'Stack',
    'Swapaxes',
    'Tile',
    'Transpose',
    'TransposeSequence',
    'Vstack',
    'Where',

    # Connection
    'Convolution2DFunction',
    'ConvolutionND',
    'Deconvolution2DFunction',
    'DeconvolutionND',
    'EmbedIDFunction',
    'LinearFunction',

    # Loss
    'SoftmaxCrossEntropy',

    # Math
    'Absolute',
    'Add',
    'AddConstant',
    'Arccos',
    'Arcsin',
    'Arctan',
    'ArgMax',
    'ArgMin',
    'BroadcastTo',
    'Clip',
    'Cos',
    'Cosh',
    'Div',
    'DivFromConstant',
    'Exp',
    'Identity',
    'LinearInterpolate',
    'Log',
    'LogSumExp',
    'MatMul',
    'Max',
    'Maximum',
    'Mean',
    'Min',
    'Minimum',
    'Mul',
    'MulConstant',
    'Neg',
    'PowConstVar',
    'PowVarConst',
    'PowVarVar',
    'Prod',
    'RsqrtGPU',
    'sign',
    'Sin',
    'Sinh',
    'Sqrt',
    'Square',
    'Sub',
    'SubFromConstant',
    'Sum',
    'Tan',

    # Noise
    'Dropout',

    # Normalization
    'BatchNormalization',
    'FixedBatchNormalization',
    'GroupNormalization',
    'LocalResponseNormalization',
    'NormalizeL2',

    # Pooling
    'AveragePooling2D',
    'AveragePoolingND',
    'MaxPooling2D',
    'MaxPoolingND',
    'ROIPooling2D',
    'Unpooling2D',

    # RNN
    'n_step_gru',
}

_converters = None


def _get_converters():
    global _converters

    if _converters is not None:
        return _converters

    _converters = {
        name: FunctionConverter(getattr(functions, 'convert_'+name, None))
        for name in _supported_function_node_set}
    return _converters


converters = _get_converters()


_supported_function_set = {
    # Math
    (F, 'sign'),

    # RNN
    (F, 'n_step_gru'),
    (F.rnn.n_step_gru, 'n_step_gru'),
}


@contextmanager
def patch_functions():
    org_funcs = {}
    for mod, name in _supported_function_set:
        org_func = getattr(mod, name)
        org_funcs[(mod, name)] = org_func
        setattr(mod, name, fake_as_funcnode(
            org_func, name, experimental_warning=False))
    try:
        yield
    finally:
        for mod, name in _supported_function_set:
            setattr(mod, name, org_funcs[(mod, name)])



from chainer.functions.rnn import n_step_lstm as rnn
from chainer.links.rnn import n_step_rnn


class NStepLSTMBase(n_step_rnn.NStepRNNBase):
    """Base link class for Stacked LSTM/BiLSTM links.

    This link is base link class for :func:`chainer.links.NStepLSTM` and
    :func:`chainer.links.NStepBiLSTM`.

    This link's behavior depends on argument, ``use_bi_direction``.

    Args:
        n_layers (int): Number of layers.
        in_size (int): Dimensionality of input vectors.
        out_size (int): Dimensionality of hidden states and output vectors.
        dropout (float): Dropout ratio.
        use_bi_direction (bool): if ``True``, use Bi-directional LSTM.

    .. seealso::
        :func:`chainer.functions.n_step_lstm`
        :func:`chainer.functions.n_step_bilstm`

    """

    n_weights = 8

    def forward(self, hx, cx, xs, **kwargs):
        """forward(self, hx, cx, xs)

        Calculates all of the hidden states and the cell states.

        Args:
            hx (:class:`~chainer.Variable` or None):
                Initial hidden states. If ``None`` is specified zero-vector
                is used. Its shape is ``(S, B, N)`` for uni-directional LSTM
                and ``(2S, B, N)`` for bi-directional LSTM where ``S`` is
                the number of layers and is equal to ``n_layers``,
                ``B`` is the mini-batch size,
                and ``N`` is the dimension of the hidden units.
            cx (:class:`~chainer.Variable` or None):
                Initial cell states. If ``None`` is specified zero-vector is
                used.  It has the same shape as ``hx``.
            xs (list of :class:`~chainer.Variable`): List of input sequences.
                Each element ``xs[i]`` is a :class:`chainer.Variable` holding
                a sequence. Its shape is ``(L_i, I)``, where ``L_i`` is the
                length of a sequence for batch ``i``, and ``I`` is the size of
                the input and is equal to ``in_size``.

        Returns:
            tuple: This function returns a tuple containing three elements,
            ``hy``, ``cy`` and ``ys``.

            - ``hy`` is an updated hidden states whose shape is the same as
              ``hx``.
            - ``cy`` is an updated cell states whose shape is the same as
              ``cx``.
            - ``ys`` is a list of :class:`~chainer.Variable` . Each element
              ``ys[i]`` holds hidden states of the last layer corresponding
              to an input ``xs[i]``. Its shape is ``(L_i, N)`` for
              uni-directional LSTM and ``(L_i, 2N)`` for bi-directional LSTM
              where ``L_i`` is the length of a sequence for batch ``i``,
              and ``N`` is size of hidden units.
        """
        (hy, cy), ys = self._call([hx, cx], xs, **kwargs)
        return hy, cy, ys


class NStepLSTM(NStepLSTMBase):
    """__init__(self, n_layers, in_size, out_size, dropout)

    Stacked Uni-directional LSTM for sequences.

    This link is stacked version of Uni-directional LSTM for sequences.
    It calculates hidden and cell states of all layer at end-of-string,
    and all hidden states of the last layer for each time.

    Unlike :func:`chainer.functions.n_step_lstm`, this function automatically
    sort inputs in descending order by length, and transpose the sequence.
    Users just need to call the link with a list of :class:`chainer.Variable`
    holding sequences.

    Args:
        n_layers (int): Number of layers.
        in_size (int): Dimensionality of input vectors.
        out_size (int): Dimensionality of hidden states and output vectors.
        dropout (float): Dropout ratio.

    .. seealso::
        :func:`chainer.functions.n_step_lstm`

    .. admonition:: Example

        *Read* :meth:`forward` *method below first.*

        >>> dropout_ratio = 0.0
        >>> in_size, seq_len, n_layers, out_size = 2, 4, 2, 3
        >>> batch = 5
        >>> xs = [
        ...     Variable(np.random.rand(seq_len, in_size).astype(np.float32))
        ...     for i in range(batch)]
        >>> [x.shape for x in xs]
        [(4, 2), (4, 2), (4, 2), (4, 2), (4, 2)]
        >>> lstm = L.NStepLSTM(n_layers, in_size, out_size, dropout_ratio)

        Without hidden or cell state:

        >>> hy, cy, ys = lstm(None, None, xs)
        >>> hy.shape  # shape should be (n_layers, batch, out_size)
        (2, 5, 3)
        >>> ys[0].shape  # should be (seq_len, out_size)
        (4, 3)
        >>> len(ys)  # should be equal to batch
        5

        With hidden and cell states:

        >>> h_shape = (n_layers, batch, out_size)
        >>> hx = Variable(np.ones(h_shape, np.float32))
        >>> cx = Variable(np.ones(h_shape, np.float32))
        >>> hy, cy, ys = lstm(hx, cx, xs)
        >>> hy.shape  # shape should be (n_layers, batch, out_size)
        (2, 5, 3)
        >>> ys[0].shape  # should be (seq_len, out_size)
        (4, 3)

    """

    use_bi_direction = False

    def rnn(self, *args):
        return rnn.n_step_lstm(*args)

    @property
    def n_cells(self):
        return 2


class NStepBiLSTM(NStepLSTMBase):
    """__init__(self, n_layers, in_size, out_size, dropout)

    Stacked Bi-directional LSTM for sequences.

    This link is stacked version of Bi-directional LSTM for sequences.
    It calculates hidden and cell states of all layer at end-of-string,
    and all hidden states of the last layer for each time.

    Unlike :func:`chainer.functions.n_step_bilstm`, this function automatically
    sort inputs in descending order by length, and transpose the sequence.
    Users just need to call the link with a list of :class:`chainer.Variable`
    holding sequences.

    Args:
        n_layers (int): Number of layers.
        in_size (int): Dimensionality of input vectors.
        out_size (int): Dimensionality of hidden states and output vectors.
        dropout (float): Dropout ratio.

    .. seealso::
        :func:`chainer.functions.n_step_bilstm`

    """

    use_bi_direction = True

    def rnn(self, *args):
        return rnn.n_step_bilstm(*args)

    @property
    def n_cells(self):
        return 2
